/*  This example sketch is for the Botletics SIM7000/7070 shield and Arduino
 *  to collect GPS, temperature, and battery data and send those values via MQTT
 *  to just about any MQTT broker.
 *  
 *  NOTE: MQTTS development is in progress
 *  
 *  Just make sure to replace credentials with your own, and change the names of the
 *  topics you want to publish or subscribe to.
 *  
 *  Author: Timothy Woo (www.botletics.com)
 *  Github: https://github.com/botletics/SIM7000-LTE-Shield
 *  Last Updated: 7/4/2022
 *  License: GNU GPL v3.0
 */

#include "BotleticsSIM7000.h" // https://github.com/botletics/SIM7000-Shield-Library/tree/master/src

#if defined(ARDUINO_SAMD_ZERO) && defined(SERIAL_PORT_USBVIRTUAL)
  // Required for Serial on Zero based boards
  #define Serial SERIAL_PORT_USBVIRTUAL
#endif

/************************* PIN DEFINITIONS *********************************/
// For botletics SIM70X0 shield
#define BOTLETICS_PWRKEY 6
#define RST 7
//#define DTR 8 // Connect with solder jumper
//#define RI 9 // Need to enable via AT commands
#define TX 10 // Microcontroller RX
#define RX 11 // Microcontroller TX
//#define T_ALERT 12 // Connect with solder jumper

#define LED 13 // Just for testing if needed!

#define samplingRate 30 // The time we want to delay after each post (in seconds)

// We default to using software serial. If you want to use hardware serial
// (because softserial isnt supported) comment out the following three lines 
// and uncomment the HardwareSerial line
#include <SoftwareSerial.h>
SoftwareSerial modemSS = SoftwareSerial(TX, RX);

// Use the following line for ESP8266 instead of the line above (comment out the one above)
//SoftwareSerial modemSS = SoftwareSerial(TX, RX, false, 256); // TX, RX, inverted logic, buffer size

SoftwareSerial *modemSerial = &modemSS;

// Hardware serial is also possible!
//HardwareSerial *modemSerial = &Serial1;

// For ESP32 hardware serial use these lines instead
//#include <HardwareSerial.h>
//HardwareSerial modemSS(1);

Botletics_modem_LTE modem = Botletics_modem_LTE();

/************************* MQTT PARAMETERS *********************************/
#define MQTT_SERVER      "m10.cloudmqtt.com"
#define MQTT_PORT        16644
#define MQTT_USERNAME    "MQTT_USERNAME"
#define MQTT_PASSWORD    "MQTT_PASSWORD"

// Set topic names to publish and subscribe to
#define GPS_TOPIC       "location"
#define TEMP_TOPIC      "temperature"
#define BATT_TOPIC      "battery"
#define SUB_TOPIC       "command"     // Subscribe topic name

// How many transmission failures in a row we're OK with before reset
uint8_t txfailures = 0;

/****************************** OTHER STUFF ***************************************/
// For temperature sensor
#include <Wire.h>
#include "Adafruit_MCP9808.h"

// Create the MCP9808 temperature sensor object
Adafruit_MCP9808 tempsensor = Adafruit_MCP9808();

uint8_t readline(char *buff, uint8_t maxbuff, uint16_t timeout = 0);
char imei[16] = {0}; // Use this for device ID
uint8_t type;
uint16_t battLevel = 0; // Battery level (percentage)
float latitude, longitude, speed_kph, heading, altitude, second;
uint16_t year;
uint8_t month, day, hour, minute;
uint8_t counter = 0;
unsigned long timer = 0;
bool firstTime = true;
//char PIN[5] = "1234"; // SIM card PIN

// NOTE: Keep the buffer sizes as small as possible, espeially on
// Arduino Uno which doesn't have much computing power to handle
// large buffers. On Arduino Mega you shouldn't have to worry much.
char latBuff[12], longBuff[12], locBuff[50], speedBuff[12],
     headBuff[12], altBuff[12], tempBuff[12], battBuff[12];

char replybuffer[255]; // For reading stuff coming through UART, like subscribed topic messages

void setup() {
  Serial.begin(9600);
  Serial.println(F("*** SIM70X0 MQTT Example ***"));

  #ifdef LED
    pinMode(LED, OUTPUT);
    digitalWrite(LED, LOW);
  #endif
  
  pinMode(RST, OUTPUT);
  digitalWrite(RST, HIGH); // Default state

  modem.powerOn(BOTLETICS_PWRKEY); // Power on the module
  moduleSetup(); // Establishes first-time serial comm and prints IMEI

  if (!tempsensor.begin()) {
    Serial.println("Couldn't find the MCP9808!");
    tempsensor.wake(); // Wake up the MCP9808 if it was sleeping and retry
    if (!tempsensor.begin()) while (1);
  }

  // Unlock SIM card if needed
  // Remember to uncomment the "PIN" variable definition above
  /*
  if (!modem.unlockSIM(PIN)) {
    Serial.println(F("Failed to unlock SIM card"));
  }
  */

  // Set modem to full functionality
  modem.setFunctionality(1); // AT+CFUN=1

  // Configure a GPRS APN, username, and password.
  // You might need to do this to access your network's GPRS/data
  // network.  Contact your provider for the exact APN, username,
  // and password values.  Username and password are optional and
  // can be removed, but APN is required.
  //modem.setNetworkSettings(F("your APN"), F("your username"), F("your password"));
  //modem.setNetworkSettings(F("m2m.com.attz")); // For AT&T IoT SIM card
  //modem.setNetworkSettings(F("telstra.internet")); // For Telstra (Australia) SIM card - CAT-M1 (Band 28)
  modem.setNetworkSettings(F("hologram")); // For Hologram SIM card

  // Optionally configure HTTP gets to follow redirects over SSL.
  // Default is not to follow SSL redirects, however if you uncomment
  // the following line then redirects over SSL will be followed.
  //modem.setHTTPSRedirect(true);

  /*
  // Other examples of some things you can set:
  modem.setPreferredMode(38); // Use LTE only, not 2G
  modem.setPreferredLTEMode(1); // Use LTE CAT-M only, not NB-IoT
  modem.setOperatingBand("CAT-M", 12); // AT&T uses band 12
//  modem.setOperatingBand("CAT-M", 13); // Verizon uses band 13
  modem.enableRTC(true);
  
  modem.enableSleepMode(true);
  modem.set_eDRX(1, 4, "0010");
  modem.enablePSM(true);

  // Set the network status LED blinking pattern while connected to a network (see AT+SLEDS command)
  modem.setNetLED(true, 2, 64, 3000); // on/off, mode, timer_on, timer_off
  modem.setNetLED(false); // Disable network status LED
  */

  // Perform first-time GPS/data setup if the shield is going to remain on,
  // otherwise these won't be enabled in loop() and it won't work!
  // Enable GPS
  while (!modem.enableGPS(true)) {
    Serial.println(F("Failed to turn on GPS, retrying..."));
    delay(2000); // Retry every 2s
  }
  Serial.println(F("Turned on GPS!"));
}

void loop() {
  if (firstTime || millis() - timer > samplingRate * 1000UL) {    
    // Connect to cell network and verify connection
    // If unsuccessful, keep retrying every 2s until a connection is made
    while (!netStatus()) {
      Serial.println(F("Failed to connect to cell network, retrying..."));
      delay(2000); // Retry every 2s
    }
    Serial.println(F("Connected to cell network!"));
  
    // Disable data just to make sure it was actually off so that we can turn it on
  //  modem.openWirelessConnection(false);
  
    // Open wireless connection if not already activated
    if (!modem.wirelessConnStatus()) {
      while (!modem.openWirelessConnection(true)) {
        Serial.println(F("Failed to enable connection, retrying..."));
        delay(2000); // Retry every 2s
      }
      Serial.println(F("Enabled data!"));
    }
    else {
      Serial.println(F("Data already enabled!"));
    }
  
    // Measure battery level
    battLevel = readVcc(); // Get voltage in mV
  
    // Measure temperature
    tempsensor.wake(); // Wake up the MCP9808 if it was sleeping
    float tempC = tempsensor.readTempC();
    float tempF = tempC * 9.0 / 5.0 + 32;
    Serial.print("Temp: "); Serial.print(tempC); Serial.print("*C\t"); 
    Serial.print(tempF); Serial.println("*F");
    
    Serial.println("Shutting down the MCP9808...");
    tempsensor.shutdown(); // In this mode the MCP9808 draws only about 0.1uA
  
    float temperature = tempC; // Select what unit you want to use for this example
  
    // Turn on GPS if it wasn't on already (e.g., if the module wasn't turned off)
  #ifdef turnOffShield
    while (!modem.enableGPS(true)) {
      Serial.println(F("Failed to turn on GPS, retrying..."));
      delay(2000); // Retry every 2s
    }
    Serial.println(F("Turned on GPS!"));
  #endif
  
    // Get a fix on location, try every 2s
    // Use the top line if you want to parse UTC time data as well, the line below it if you don't care
  //  while (!modem.getGPS(&latitude, &longitude, &speed_kph, &heading, &altitude, &year, &month, &day, &hour, &minute, &second)) {
    while (!modem.getGPS(&latitude, &longitude, &speed_kph, &heading, &altitude)) {
      Serial.println(F("Failed to get GPS location, retrying..."));
      delay(2000); // Retry every 2s
    }
    Serial.println(F("Found 'eeeeem!"));
    Serial.println(F("---------------------"));
    Serial.print(F("Latitude: ")); Serial.println(latitude, 6);
    Serial.print(F("Longitude: ")); Serial.println(longitude, 6);
    Serial.print(F("Speed: ")); Serial.println(speed_kph);
    Serial.print(F("Heading: ")); Serial.println(heading);
    Serial.print(F("Altitude: ")); Serial.println(altitude);
    /*
    // Uncomment this if you care about parsing UTC time
    Serial.print(F("Year: ")); Serial.println(year);
    Serial.print(F("Month: ")); Serial.println(month);
    Serial.print(F("Day: ")); Serial.println(day);
    Serial.print(F("Hour: ")); Serial.println(hour);
    Serial.print(F("Minute: ")); Serial.println(minute);
    Serial.print(F("Second: ")); Serial.println(second);
    */
    Serial.println(F("---------------------"));
  
    // Format the floating point numbers
    dtostrf(latitude, 1, 6, latBuff); // float_val, min_width, digits_after_decimal, char_buffer
    dtostrf(longitude, 1, 6, longBuff);
    dtostrf(speed_kph, 1, 0, speedBuff);
    dtostrf(heading, 1, 0, headBuff);
    dtostrf(altitude, 1, 1, altBuff);
    dtostrf(temperature, 1, 2, tempBuff);
    dtostrf(battLevel, 1, 0, battBuff);
  
    // Construct a combined, comma-separated location array
    sprintf(locBuff, "%s,%s,%s,%s", speedBuff, latBuff, longBuff, altBuff); // This could look like "10,33.123456,-85.123456,120.5"
    
    // If not already connected, connect to MQTT
    if (! modem.MQTT_connectionStatus()) {
      // Set up MQTT parameters (see MQTT app note for explanation of parameter values)
      modem.MQTT_setParameter("URL", MQTT_SERVER, MQTT_PORT);
      // Set up MQTT username and password if necessary
      modem.MQTT_setParameter("USERNAME", MQTT_USERNAME);
      modem.MQTT_setParameter("PASSWORD", MQTT_PASSWORD);
  //    modem.MQTTsetParameter("KEEPTIME", 30); // Time to connect to server, 60s by default
      
      Serial.println(F("Connecting to MQTT broker..."));
      if (! modem.MQTT_connect(true)) {
        Serial.println(F("Failed to connect to broker!"));
      }
    }
    else {
      Serial.println(F("Already connected to MQTT server!"));
    }
  
    // Now publish all the GPS and temperature data to their respective topics!
    // Parameters for MQTT_publish: Topic, message (0-512 bytes), message length, QoS (0-2), retain (0-1)
    if (!modem.MQTT_publish(GPS_TOPIC, locBuff, strlen(locBuff), 1, 0)) Serial.println(F("Failed to publish!")); // Send GPS location
    if (!modem.MQTT_publish(TEMP_TOPIC, tempBuff, strlen(tempBuff), 1, 0)) Serial.println(F("Failed to publish!")); // Send temperature
    if (!modem.MQTT_publish(BATT_TOPIC, battBuff, strlen(battBuff), 1, 0)) Serial.println(F("Failed to publish!")); // Send battery level
  
    // Note the command below may error out if you're already subscribed to the topic!
    modem.MQTT_subscribe(SUB_TOPIC, 1); // Topic name, QoS
    
    // Unsubscribe from topics if wanted:
  //  modem.MQTT_unsubscribe(SUB_TOPIC);
  
    // Enable MQTT data format to hex
  //  modem.MQTT_dataFormatHex(true); // Input "false" to reverse
  
    // Disconnect from MQTT
  //  modem.MQTT_connect(false);
  
    // Delay until next post but read incoming subscribed topic messages (if any)
    Serial.print(F("Waiting for ")); Serial.print(samplingRate); Serial.println(F(" seconds\r\n"));

    firstTime = false;
    timer = millis(); // Reset timer at the end
  }
  else {
    // The rest of the time, read anything coming over via UART from the SIM7000
    // If it's from an MQTT subscribed topic message, parse it
    uint8_t i = 0;
    if (modem.available()) {
      while (modem.available()) {
        replybuffer[i] = modem.read();
        i++;
      }

      Serial.print(replybuffer); // DEBUG
      delay(100); // Make sure it prints and also allow other stuff to run properly

      // We got an MQTT message! Parse the topic and message
      // Format: +SMSUB: "topic_name","message"
      if (strstr(replybuffer, "+SMSUB:") != NULL) {
        Serial.println(F("*** Received MQTT message! ***"));
        
        char *p = strtok(replybuffer, ",\"");
        char *topic_p = strtok(NULL, ",\"");
        char *message_p = strtok(NULL, ",\"");
        
        Serial.print(F("Topic: ")); Serial.println(topic_p);
        Serial.print(F("Message: ")); Serial.println(message_p);
  
        // Do something with the message
        // For example, if the topic was "command" and we received a "yes", turn on an LED!
        if (strcmp(topic_p, "command") == 0) {
          if (strcmp(message_p, "yes") == 0) {
            Serial.println(F("Turning on LED!"));
            digitalWrite(LED, HIGH);
          }
          else if (strcmp(message_p, "no") == 0) {
            Serial.println(F("Turning off LED!"));
            digitalWrite(LED, HIGH);
          }
        }
      }

      /*
      // Alternatively, could convert to String class and parse that way
      // Format: +SMSUB: "topic_name","message"
      String reply = String(replybuffer);
      Serial.println(reply);
      
      if (reply.indexOf("+SMSUB: ") != -1) {
        Serial.println(F("*** Received MQTT message! ***"));
        // Chop off the "SMSUB: " part plus the beginning quote
        // After this, reply should be: "topic_name","message"
        reply = reply.substring(9);
        uint8_t idx = reply.indexOf("\",\""); // Search for second quote
        String topic = reply.substring(1, idx); // Grab only the text (without quotes)
        String message = reply.substring(idx+3, reply.length()-3);
        
        Serial.print(F("Topic: ")); Serial.println(topic);
        Serial.print(F("Message: ")); Serial.println(message);
        // Do something with the message
        // For example, if the topic was "command" and we received a "yes", turn on an LED!
        if (topic == "command") {
          if (message == "yes") {
            Serial.println(F("Turning on LED!"));
            digitalWrite(LED, HIGH);
          }
          else if (message == "no") {
            Serial.println(F("Turning off LED!"));
            digitalWrite(LED, HIGH);
          }
        }
      }
      */
    }
  }
}

void moduleSetup() {
  // SIM7000 takes about 3s to turn on and SIM7500 takes about 15s
  // Press Arduino reset button if the module is still turning on and the board doesn't find it.
  // When the module is on it should communicate right after pressing reset

  // Software serial:
  modemSS.begin(115200); // Default SIM7000 shield baud rate

  Serial.println(F("Configuring to 9600 baud"));
  modemSS.println("AT+IPR=9600"); // Set baud rate
  delay(100); // Short pause to let the command run
  modemSS.begin(9600);
  if (! modem.begin(modemSS)) {
    Serial.println(F("Couldn't find modem"));
    while (1); // Don't proceed if it couldn't find the device
  }

  // Hardware serial:
  /*
  modemSerial->begin(115200); // Default SIM7000 baud rate

  if (! modem.begin(*modemSerial)) {
    DEBUG_PRINTLN(F("Couldn't find SIM7000"));
  }
  */

  // The commented block of code below is an alternative that will find the module at 115200
  // Then switch it to 9600 without having to wait for the module to turn on and manually
  // press the reset button in order to establish communication. However, once the baud is set
  // this method will be much slower.
  /*
  modemSerial->begin(115200); // Default LTE shield baud rate
  modem.begin(*modemSerial); // Don't use if statement because an OK reply could be sent incorrectly at 115200 baud

  Serial.println(F("Configuring to 9600 baud"));
  modem.setBaudrate(9600); // Set to 9600 baud
  modemSerial->begin(9600);
  if (!modem.begin(*modemSerial)) {
    Serial.println(F("Couldn't find modem"));
    while(1); // Don't proceed if it couldn't find the device
  }
  */

  type = modem.type();
  Serial.println(F("Modem is OK"));
  Serial.print(F("Found "));
  switch (type) {
    case SIM800L:
      Serial.println(F("SIM800L")); break;
    case SIM800H:
      Serial.println(F("SIM800H")); break;
    case SIM808_V1:
      Serial.println(F("SIM808 (v1)")); break;
    case SIM808_V2:
      Serial.println(F("SIM808 (v2)")); break;
    case SIM5320A:
      Serial.println(F("SIM5320A (American)")); break;
    case SIM5320E:
      Serial.println(F("SIM5320E (European)")); break;
    case SIM7000:
      Serial.println(F("SIM7000")); break;
    case SIM7070:
      Serial.println(F("SIM7070")); break;
    case SIM7500:
      Serial.println(F("SIM7500")); break;
    case SIM7600:
      Serial.println(F("SIM7600")); break;
    default:
      Serial.println(F("???")); break;
  }
  
  // Print module IMEI number.
  uint8_t imeiLen = modem.getIMEI(imei);
  if (imeiLen > 0) {
    Serial.print("Module IMEI: "); Serial.println(imei);
  }
}

// Read the module's power supply voltage
float readVcc() {
  // Read battery voltage
  if (!modem.getBattVoltage(&battLevel)) Serial.println(F("Failed to read batt"));
  else Serial.print(F("battery = ")); Serial.print(battLevel); Serial.println(F(" mV"));

  // Read LiPo battery percentage
//  if (!modem.getBattPercent(&battLevel)) Serial.println(F("Failed to read batt"));
//  else Serial.print(F("BAT % = ")); Serial.print(battLevel); Serial.println(F("%"));

  return battLevel;
}

bool netStatus() {
  int n = modem.getNetworkStatus();
  
  Serial.print(F("Network status ")); Serial.print(n); Serial.print(F(": "));
  if (n == 0) Serial.println(F("Not registered"));
  if (n == 1) Serial.println(F("Registered (home)"));
  if (n == 2) Serial.println(F("Not registered (searching)"));
  if (n == 3) Serial.println(F("Denied"));
  if (n == 4) Serial.println(F("Unknown"));
  if (n == 5) Serial.println(F("Registered roaming"));

  if (!(n == 1 || n == 5)) return false;
  else return true;
}
